// ///////////////////  ***********************  \\\\\\\\\\\\\\\\\\\\ //
//
//  SoGlutRenderArea, version 2.0
//
//  a simple window component like SoXtRenderArea for Inventor
//
//  Copyright (C) 2001-2002 by PC John. All rights reserved.
//
//  This game is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public License
//  version 2.1 as published by the Free Software Foundation. See the
//  file LICENSE.LGPL at the root directory of the distribution for
//  more details.
//
//  If you want to use Sector 5 code for applications not compatible 
//  with the LGPL license, please contact PC John at pcjohn@email.cz.
//
// \\\\\\\\\\\\\\\\\\\  ***********************  //////////////////// //


// Windows specific
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
# include <windows.h>
# undef DELETE // Microsoft defines it and it collides with SoKeyboardEvent::DELETE

// to avoid Microsoft Visual Studio to generate warnings when
// compiling Coin's headers if dll-linking
# pragma warning(disable:4275)
# pragma warning(disable:4251)
#endif

#include <Inventor/SoDB.h>
#include <Inventor/SoInteraction.h>
#include <Inventor/nodekits/SoNodeKit.h>
#include <Inventor/fields/SoSFTime.h>
#include <Inventor/elements/SoGLCacheContextElement.h>
#include <Inventor/errors/SoDebugError.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "SoGlutRenderArea.h"


#ifndef SO_GLUT_USE_GAME_MODE
# ifdef __APPLE__  // game mode does not work properly on Macs
#  define SO_GLUT_USE_GAME_MODE 0 
# else
#  define SO_GLUT_USE_GAME_MODE 1
# endif
#endif


static char SO_GLUT_DEFAULT_WINDOW_TITLE[] = "SoGlutRenderArea";


SoGlutRenderArea* SoGlut::windowsList = NULL;
bool SoGlut::widgetVisible = false;
static SoGlutRenderArea::Resolution *availableRes = NULL;
static bool fullScreenActive = false;



static void glut_redraw_cb(void *user, SoSceneManager *manager);
static void glut_expose_cb(void);
static void glut_reshape_cb(int w, int h);
static void glut_idle_cb(void);
static void glut_mouse_motion_cb(int x, int y);
static void glut_mouse_button_cb(int button, int state, int x, int y);
static void glut_keyboard_cb(unsigned char key, int x, int y);
static void glut_keyboard_up_cb(unsigned char key, int x, int y);
static void glut_speckeyboard_cb(int key, int x, int y);
static void glut_speckeyboard_up_cb(int key, int x, int y);
static void sensorQueueChanged_CB(void *data);


// Ascii to SoKeyboardEvent::Key translation table for first 128 characters
#define Xx SoKeyboardEvent::ANY // many characters can not be translated
static const SoKeyboardEvent::Key keyTranslateTable[128] =
{
  Xx, Xx, Xx, Xx, Xx, Xx, Xx, Xx, // 0-7
  SoKeyboardEvent::BACKSPACE, // 8
  SoKeyboardEvent::TAB, // 9
  Xx, Xx, Xx, // 10-12
  SoKeyboardEvent::ENTER, // 13
  Xx, Xx, Xx, Xx, Xx, // 14-18
  SoKeyboardEvent::PAUSE, // 19
  Xx, Xx, Xx, Xx, Xx, Xx, Xx, // 20-26
  SoKeyboardEvent::ESCAPE, // 27
  Xx, Xx, // 28-29
  Xx, Xx, // 30-31

  SoKeyboardEvent::SPACE, // 32
  Xx, Xx, Xx, Xx, Xx, Xx, // 33-38
  SoKeyboardEvent::APOSTROPHE, // 39 "'"
  Xx, Xx, Xx, Xx, // 40-43

  SoKeyboardEvent::COMMA, // 44 ","
  SoKeyboardEvent::MINUS, // 45 "-"
  SoKeyboardEvent::PERIOD, // 46 "."
  SoKeyboardEvent::SLASH, // 47 "/"
  
  SoKeyboardEvent::NUMBER_0, // 48
  SoKeyboardEvent::NUMBER_1,
  SoKeyboardEvent::NUMBER_2,
  SoKeyboardEvent::NUMBER_3,
  SoKeyboardEvent::NUMBER_4,
  SoKeyboardEvent::NUMBER_5,
  SoKeyboardEvent::NUMBER_6,
  SoKeyboardEvent::NUMBER_7,
  SoKeyboardEvent::NUMBER_8,
  SoKeyboardEvent::NUMBER_9, // 57

  Xx, // 58
  SoKeyboardEvent::SEMICOLON, // 59 ";"
  Xx, // 60
  SoKeyboardEvent::EQUAL, // 61 "="
  Xx, Xx, Xx, // 62-64

  SoKeyboardEvent::A, // 65 "A"
  SoKeyboardEvent::B,
  SoKeyboardEvent::C,
  SoKeyboardEvent::D,
  SoKeyboardEvent::E,
  SoKeyboardEvent::F,
  SoKeyboardEvent::G,
  SoKeyboardEvent::H,
  SoKeyboardEvent::I,
  SoKeyboardEvent::J,
  SoKeyboardEvent::K,
  SoKeyboardEvent::L,
  SoKeyboardEvent::M,
  SoKeyboardEvent::N,
  SoKeyboardEvent::O,
  SoKeyboardEvent::P,
  SoKeyboardEvent::Q,
  SoKeyboardEvent::R,
  SoKeyboardEvent::S,
  SoKeyboardEvent::T,
  SoKeyboardEvent::U,
  SoKeyboardEvent::V,
  SoKeyboardEvent::W,
  SoKeyboardEvent::X,
  SoKeyboardEvent::Y,
  SoKeyboardEvent::Z, // 90 "Z"
  
  SoKeyboardEvent::BRACKETLEFT, // 91 "["
  SoKeyboardEvent::BACKSLASH, // 92 "\"
  SoKeyboardEvent::BRACKETRIGHT, // 93 "]"
  Xx, Xx, // 94-95
  SoKeyboardEvent::GRAVE, // 96 "`"

  SoKeyboardEvent::A, // 97 "a"
  SoKeyboardEvent::B,
  SoKeyboardEvent::C,
  SoKeyboardEvent::D,
  SoKeyboardEvent::E,
  SoKeyboardEvent::F,
  SoKeyboardEvent::G,
  SoKeyboardEvent::H,
  SoKeyboardEvent::I,
  SoKeyboardEvent::J,
  SoKeyboardEvent::K,
  SoKeyboardEvent::L,
  SoKeyboardEvent::M,
  SoKeyboardEvent::N,
  SoKeyboardEvent::O,
  SoKeyboardEvent::P,
  SoKeyboardEvent::Q,
  SoKeyboardEvent::R,
  SoKeyboardEvent::S,
  SoKeyboardEvent::T,
  SoKeyboardEvent::U,
  SoKeyboardEvent::V,
  SoKeyboardEvent::W,
  SoKeyboardEvent::X,
  SoKeyboardEvent::Y,
  SoKeyboardEvent::Z, // 122 "z"

  Xx, Xx, Xx, Xx, // 123-126

  (SoKeyboardEvent::Key)0xffff // To avoid all troubles with DELETE
                               // key we use this instead of simple
                               // SoKeyboardEvent::DELETE
};

// GLUT gives no access to these keys:
//
// SoKeyboardEvent::LEFT_SHIFT
// SoKeyboardEvent::RIGHT_SHIFT
// SoKeyboardEvent::LEFT_CONTROL
// SoKeyboardEvent::RIGHT_CONTROL
// SoKeyboardEvent::LEFT_ALT
// SoKeyboardEvent::RIGHT_ALT

// SoKeyboardEvent::PRIOR
// SoKeyboardEvent::NEXT
// SoKeyboardEvent::PAD_ENTER
// SoKeyboardEvent::PAD_F1
// SoKeyboardEvent::PAD_F2
// SoKeyboardEvent::PAD_F3
// SoKeyboardEvent::PAD_F4
// SoKeyboardEvent::PAD_0
// SoKeyboardEvent::PAD_1
// SoKeyboardEvent::PAD_2
// SoKeyboardEvent::PAD_3
// SoKeyboardEvent::PAD_4
// SoKeyboardEvent::PAD_5
// SoKeyboardEvent::PAD_6
// SoKeyboardEvent::PAD_7
// SoKeyboardEvent::PAD_8
// SoKeyboardEvent::PAD_9
// SoKeyboardEvent::PAD_ADD
// SoKeyboardEvent::PAD_SUBTRACT
// SoKeyboardEvent::PAD_MULTIPLY
// SoKeyboardEvent::PAD_DIVIDE
// SoKeyboardEvent::PAD_SPACE
// SoKeyboardEvent::PAD_TAB
// SoKeyboardEvent::PAD_INSERT
// SoKeyboardEvent::PAD_DELETE
// SoKeyboardEvent::PAD_PERIOD

// SoKeyboardEvent::RETURN - same as SoKeyboardEvent::ENTER
// SoKeyboardEvent::SCROLL_LOCK

// SoKeyboardEvent::PRINT
// SoKeyboardEvent::NUM_LOCK
// SoKeyboardEvent::CAPS_LOCK
// SoKeyboardEvent::SHIFT_LOCK


// keyboard's Fx to SoKeyboardEvent::Key translation table
static const SoKeyboardEvent::Key specialKeyTranslateTable1[] =
{
  SoKeyboardEvent::F1,
  SoKeyboardEvent::F2,
  SoKeyboardEvent::F3,
  SoKeyboardEvent::F4,
  SoKeyboardEvent::F5,
  SoKeyboardEvent::F6,
  SoKeyboardEvent::F7,
  SoKeyboardEvent::F8,
  SoKeyboardEvent::F9,
  SoKeyboardEvent::F10,
  SoKeyboardEvent::F11,
  SoKeyboardEvent::F12
};

// keyboard's special characters to SoKeyboardEvent::Key translation table
static const SoKeyboardEvent::Key specialKeyTranslateTable2[] =
{
  SoKeyboardEvent::LEFT_ARROW,
  SoKeyboardEvent::UP_ARROW,
  SoKeyboardEvent::RIGHT_ARROW,
  SoKeyboardEvent::DOWN_ARROW,
  SoKeyboardEvent::PAGE_UP,
  SoKeyboardEvent::PAGE_DOWN,
  SoKeyboardEvent::HOME,
  SoKeyboardEvent::END,
  SoKeyboardEvent::INSERT
};

#undef Xx



Widget SoGlut::init(const char *appName, const char *className)
{
  // try to make fictive command-line arguments
  char *s = (char*)malloc(strlen(appName)+1);
  strcpy(s, appName);
  char *p = &s[0];
  char **argv = &p;

  // call init
  Widget r = init(1, argv);
  free(s);
  return r;
}



Widget SoGlut::init(int argc, char ** argv)
{
  // initialize OIV
  SoDB::init();
  SoNodeKit::init();
  SoInteraction::init();

  // initialize GLUT
  glutInit(&argc, argv); // this function can terminate application
  glutIdleFunc(glut_idle_cb);
  SoDB::getSensorManager()->setChangedCallback(sensorQueueChanged_CB, NULL);

  // register atexit
  static bool atexitRegistered = false;
  if (!atexitRegistered) {
    atexitRegistered = true;
    if (atexit(SoGlut::done) != 0)
      SoDebugError::post("SoGlut::init", "Can not register atexit function.");
  }

  // return widget, must be non-zero
  return (Widget) 0xffffffff;
}



void SoGlut::done()
{
  SoGlut::hide();
  if (availableRes) {
    free(availableRes);
    availableRes = NULL;
  }
}



void SoGlut::mainLoop()
{
  glutMainLoop();
}



void SoGlut::show(Widget widget)
// shows all windows of application with visible set to true
{
  if (SoGlut::widgetVisible)  return;

  SoGlut::widgetVisible = true;

  SoGlutRenderArea *w = SoGlut::windowsList;
  while (w != NULL)
  {
    if (w->visible == true)
      w->doShow();
    w = w->nextWindowsListItem;
  }
}



void SoGlut::hide(Widget widget)
// hides all application windows
{
  if (SoGlut::widgetVisible == false)  return;

  SoGlut::widgetVisible = false;

  SoGlutRenderArea *w = SoGlut::windowsList;
  while (w != NULL)
  {
    if (w->visible == true)
      w->doHide();
    w = w->nextWindowsListItem;
  }
}



void SoGlut::addWindow(SoGlutRenderArea *window)
// adds window to the end of the list
{
  SoGlutRenderArea *w = windowsList;

  if (w==NULL)
  {
    windowsList = window;
    window->nextWindowsListItem = NULL;
    return;
  }

  while (w->nextWindowsListItem != NULL)
    w = w->nextWindowsListItem;

  w->nextWindowsListItem = window;
  window->nextWindowsListItem = NULL;
}



void SoGlut::removeWindow(SoGlutRenderArea *window)
{
  SoGlutRenderArea **w = &windowsList;

  while ((*w) != NULL)
  {
    if ((*w) == window)
    {
      (*w) = (*w)->nextWindowsListItem;
      break;
    }
    w = &(*w)->nextWindowsListItem;
  }
}


  
SoGlutRenderArea* SoGlut::getWindowFromHandle(Widget id)
{
  SoGlutRenderArea *w = windowsList;

  while (w != NULL)
  {
    if (id == w->handle)
      return w;

    w = w->nextWindowsListItem;
  }
  
  return NULL;
}



static void sensorQueueChanged_CB(void *data)
{
  glutIdleFunc(glut_idle_cb);
}



static void timedQueue_CB(int v)
{
  glut_idle_cb();
}



void glut_idle_cb(void)
{
  SoSensorManager *sm = SoDB::getSensorManager();

  sm->processTimerQueue();
  sm->processDelayQueue(TRUE);

  if (sm->isDelaySensorPending()) {
    glutIdleFunc(glut_idle_cb); 
  } else {
    SbTime eventTime;
    if (sm->isTimerSensorPending(eventTime)) { 
      // only timed events
      int t = int((eventTime.getValue() - SbTime::getTimeOfDay().getValue()) * 1000);
      unsigned int ut = (t<=0) ? 0 : (unsigned int)t;
      glutTimerFunc(ut, timedQueue_CB, 0);
      glutIdleFunc(NULL); 
    } else
      // no events
      glutIdleFunc(NULL); 
  }
}




/////////////////////////////////////////////////////////////////////




SoGlutRenderArea::SoGlutRenderArea(Widget parent,
    const char *name,
    SbBool buildInsideParent,
    SbBool getMouseInput,
    SbBool getKeyboardInput)
{
  handle = 0;
  visible = false;
  fullScreen = false;
  fsWidth = 640;
  fsHeight = 480;
  fsBPP = 0;
  title = SO_GLUT_DEFAULT_WINDOW_TITLE;
  iconTitle = SO_GLUT_DEFAULT_WINDOW_TITLE;
  mode = GLUT_DOUBLE;
  winSize = SbVec2s(300, 300);
  position = SbVec2s(0, 0);
  clearColorBuffer = true;
  clearZBuffer = true;
  autoRedraw = true;
  this->getMouseInput = getMouseInput;
  this->getKeyboardInput = getKeyboardInput;

  SoGlut::addWindow(this);

  sceneManager = new SoSceneManager;                            
  ownedSceneManager = true;
  sceneManager->setRenderCallback(glut_redraw_cb, (void *)this);
  sizeChanged(winSize[0], winSize[1]);
}



SoGlutRenderArea::~SoGlutRenderArea()
{
  if (title && title != SO_GLUT_DEFAULT_WINDOW_TITLE)  free(title);
  if (iconTitle && iconTitle != SO_GLUT_DEFAULT_WINDOW_TITLE)  free(iconTitle);
  if (ownedSceneManager) delete sceneManager;
  destroyWindow();
  SoGlut::removeWindow(this);
}



void SoGlutRenderArea::recreateWindow()
{
  if (fullScreen) {
    // entering fullscreen
#if SO_GLUT_USE_GAME_MODE
    if (!doEnterFullScreen()) {
      SoDebugError::post("SoGlutRenderArea::recreateWindow",
                         "Can not create full screen window.");
      fullScreen = false; // we will create normal window instead
    }
#else
    if (handle < 1) {
      handle = glutCreateWindow(title);
      if (handle < 1) {
        SoDebugError::post("SoGlutRenderArea::recreateWindow",
                           "Can not create window.");
        exit(EXIT_FAILURE);
        return;
      }
    }
    position = SbVec2s(glutGet(GLUT_WINDOW_X), glutGet(GLUT_WINDOW_Y));
    glutFullScreen();
#endif
  }

  if (!fullScreen) {
    if (handle >= 1) return;
    handle = glutCreateWindow(title);
    if (handle < 1) {
      SoDebugError::post("SoGlutRenderArea::recreateWindow",
                         "Can not create window.");
      exit(EXIT_FAILURE);
      return;
    }
  }

#ifdef __COIN__
  unsigned int cacheContext = SoGLCacheContextElement::getUniqueCacheContext();
#else // SGI or TGS Inventor
  static uint32_t uniqcc = 19720408;
  cacheContext = uniqcc++;
#endif
  sceneManager->getGLRenderAction()->setCacheContext(cacheContext);

  glutSetIconTitle(iconTitle);

  glutDisplayFunc(glut_expose_cb);
  glutReshapeFunc(glut_reshape_cb);

  if (getMouseInput) {
    glutMotionFunc(glut_mouse_motion_cb);
    glutPassiveMotionFunc(glut_mouse_motion_cb);
    glutMouseFunc(glut_mouse_button_cb);
  }

  if (getKeyboardInput) {
    glutKeyboardFunc(glut_keyboard_cb);
    glutKeyboardUpFunc(glut_keyboard_up_cb);
    glutSpecialFunc(glut_speckeyboard_cb);
    glutSpecialUpFunc(glut_speckeyboard_up_cb);
    glutIgnoreKeyRepeat(true);
  }
}



void SoGlutRenderArea::createWindow()
{
  if (handle < 1) {

    glutInitDisplayMode(mode | GLUT_RGBA | GLUT_DEPTH);

    if (!fullScreen) {
      // creating normal window
      glutInitWindowSize(winSize[0], winSize[1]);
      glutInitWindowPosition(position[0], position[1]);
    }

    if (fullScreen && fullScreenActive) { // can not have two full screen windows
      SoDebugError::post("SoGlutRenderArea::createWindow",
                         "Can not create fullscreen window if there is another visible fullscreen window.");
      assert(false && "Can not create fullscreen window if there is another visible fullscreen window.");
      return;
    }

    recreateWindow();
  }
}



void SoGlutRenderArea::destroyWindow()
{
  if (handle > 0) {
    position = getPosition();
    if (!fullScreen) glutDestroyWindow(handle);
    else doLeaveFullScreen();
    handle = 0;
  }
}



void SoGlutRenderArea::show()
{
  if (visible == true)  return;

  visible = true;

  if (SoGlut::widgetVisible)
    doShow();
}



void SoGlutRenderArea::doShow()
{
  if (handle < 1)
    createWindow();
  else {
    assert(!fullScreen);
    glutSetWindow(handle);
    glutShowWindow();
  }

  sceneManager->activate();
};



void SoGlutRenderArea::hide()
{
  if (visible == false)  return;

  visible = false;

  if (SoGlut::widgetVisible)
    doHide();
}



void SoGlutRenderArea::doHide()
{
  if (handle < 1)  return;

  if (!fullScreen) {
    glutSetWindow(handle);
    glutHideWindow();
  } else {
    doLeaveFullScreen();
    handle = 0;
  }

  sceneManager->deactivate();
};



void SoGlutRenderArea::setSceneManager(SoSceneManager *sm)
{
  assert(sm != NULL);

  if (sceneManager == sm) return;

  sm->setWindowSize(sceneManager->getWindowSize());

  if (ownedSceneManager) {
    delete sceneManager;
    ownedSceneManager = false;
  } else
    sceneManager->setRenderCallback(NULL, NULL);
  unsigned int cacheContext = sceneManager->getGLRenderAction()->getCacheContext();

  sceneManager = sm;
  sceneManager->setRenderCallback(glut_redraw_cb, (void *)this);
  sceneManager->getGLRenderAction()->setCacheContext(cacheContext);
  if (this->isVisible()) sceneManager->activate();
  else sceneManager->deactivate();
}



void SoGlutRenderArea::setTitle(const char *newTitle)
{
  if (title == NULL)
    if (newTitle == NULL) return;
    else;
  else
    if (newTitle == NULL) {
      free(title);
      title = NULL;
      goto update;
    } else
      if (strcmp(title, newTitle) == 0) return;
      else 
        if (title != SO_GLUT_DEFAULT_WINDOW_TITLE) free(title);

  title = (char*)malloc(strlen(newTitle)+1);
  strcpy(title, newTitle);
update:
  if (handle > 0 && !fullScreen) {
    glutSetWindow(handle);
    glutSetWindowTitle(title);
  }
};



void SoGlutRenderArea::setIconTitle(const char *newTitle)
{
  if (iconTitle == NULL)
    if (newTitle == NULL) return;
    else;
  else
    if (newTitle == NULL) {
      free(iconTitle);
      iconTitle = NULL;
      goto update;
    } else
      if (strcmp(iconTitle, newTitle) == 0) return;
      else 
        if (iconTitle != SO_GLUT_DEFAULT_WINDOW_TITLE) free(iconTitle);

  iconTitle = (char*)malloc(strlen(newTitle)+1);
  strcpy(iconTitle, newTitle);
update:
  if (handle > 0 && !fullScreen) {
    glutSetWindow(handle);
    glutSetIconTitle(iconTitle);
  }
}



void SoGlutRenderArea::setDoubleBuffer(SbBool onOrOff)
{
  if (handle > 0) { // switching can be done only before windows creation
    SoDebugError::postWarning("SoGlutRenderArea::setDoubleBuffer",
                              "Double buffering setting can not be changed after window creation.");
    return; 
  }
  if (onOrOff) mode |= GLUT_DOUBLE;
  else mode ^= GLUT_DOUBLE;
}



SbBool SoGlutRenderArea::isDoubleBuffer()  
{ 
  return mode & GLUT_DOUBLE ? TRUE : FALSE; 
}



void SoGlutRenderArea::setPosition(SbVec2s pos)
{
  if (fullScreen) { // can not set position on full screen window
    SoDebugError::postWarning("SoGlutRenderArea::setPosition",
                              "Window position can not be changed in full screen mode.");
    return; 
  }

  if (handle > 0) {
    glutSetWindow(handle);
    glutPositionWindow(pos[0], pos[1]);
  } else
    position = pos;
}



SbVec2s SoGlutRenderArea::getPosition() const
{
  if (fullScreen)
    return SbVec2s(0, 0);

  if (handle > 0) {
    glutSetWindow(handle);
    return SbVec2s(glutGet(GLUT_WINDOW_X), glutGet(GLUT_WINDOW_Y));
  } else
    return position;
}



void SoGlutRenderArea::setSize(const SbVec2s &size)
{
  if (handle > 0) {
    if (!fullScreen) {
      glutSetWindow(handle);
      glutReshapeWindow(size[0], size[1]);
    } else {
#if SO_GLUT_USE_GAME_MODE
      sizeChanged(size[0], size[1]);
      setFullScreen(true);
#endif
    }
  } else
    sizeChanged(size[0], size[1]);
};



SbVec2s SoGlutRenderArea::getSize() const
{
  return sceneManager->getWindowSize();
}



void SoGlutRenderArea::sizeChanged(int w, int h)
{
  sceneManager->setWindowSize(SbVec2s(w, h));

  if (fullScreen) {
    fsWidth = w;
    fsHeight = h;
  } else
    winSize = SbVec2s(w, h);

  if (handle > 0) {
    glutSetWindow(handle);
    glutPostRedisplay();
  }
}



void SoGlutRenderArea::reshape_cb(int w, int h)
{
  sizeChanged(w, h);
}



void glut_reshape_cb(int w, int h)
{
  SoGlutRenderArea *win = SoGlutRenderArea::getComponent(glutGetWindow());
  assert(win != NULL && "glutGetWindow returned invalid window handle");
  win->reshape_cb(w, h);
}



void SoGlutRenderArea::setViewportRegion(const SbViewportRegion &newRegion)
{
  sceneManager->setViewportRegion(newRegion);
};



const SbViewportRegion& SoGlutRenderArea::getViewportRegion() const
{ 
  return sceneManager->getViewportRegion();
}

  
  
void SoGlutRenderArea::setTransparencyType(SoGLRenderAction::TransparencyType type)
{
  sceneManager->getGLRenderAction()->setTransparencyType(type);
}



SoGLRenderAction::TransparencyType SoGlutRenderArea::getTransparencyType() const
{
  return sceneManager->getGLRenderAction()->getTransparencyType();
}



void SoGlutRenderArea::makeCurrent()  
{
  glutSetWindow(handle); 
};



void SoGlutRenderArea::redraw_cb()
{
  if (handle > 0) {
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    sceneManager->render(clearColorBuffer, clearZBuffer);

    glutSwapBuffers();
  }
}



void glut_redraw_cb(void *renderArea, SoSceneManager *manager)
{
  SoGlutRenderArea* win = (SoGlutRenderArea*)renderArea;
  
  if (win->isAutoRedraw()) {
    win->makeCurrent();
    win->redraw_cb();
  }
}



void glut_expose_cb(void)
{
  SoGlutRenderArea *win = SoGlutRenderArea::getComponent(glutGetWindow());
  assert(win != NULL && "glutGetWindow returned invalid window handle");
  win->redraw_cb();
}



void SoGlutRenderArea::makeEvent(SoEvent *event, int x, int y, int modifiers)
{
  // set whole event stuff
  event->setTime(SbTime::getTimeOfDay());
  event->setPosition(SbVec2s(x, this->getSize()[1] - y - 1));

  event->setShiftDown(modifiers & GLUT_ACTIVE_SHIFT ? TRUE : FALSE);
  event->setCtrlDown(modifiers & GLUT_ACTIVE_CTRL ? TRUE : FALSE);
  event->setAltDown(modifiers & GLUT_ACTIVE_ALT ? TRUE : FALSE);

  // process timer queue before event
  SoDB::getSensorManager()->processTimerQueue();

  // process event
  sceneManager->processEvent(event);
}



void glut_keyboard_cb(unsigned char key, int x, int y)
{
  SoGlutRenderArea *win = SoGlutRenderArea::getComponent(glutGetWindow());
  assert(win != NULL && "glutGetWindow returned invalid window handle");                        
  win->keyboard_cb(key, true, false, x, y);
}



void glut_keyboard_up_cb(unsigned char key, int x, int y)
{
  SoGlutRenderArea *win = SoGlutRenderArea::getComponent(glutGetWindow());
  assert(win != NULL && "glutGetWindow returned invalid window handle");
  win->keyboard_cb(key, false, false, x, y);
}



void glut_speckeyboard_cb(int key, int x, int y)
{
  SoGlutRenderArea *win = SoGlutRenderArea::getComponent(glutGetWindow());
  assert(win != NULL && "glutGetWindow returned invalid window handle");
  win->keyboard_cb(key, true, true, x, y);
}



void glut_speckeyboard_up_cb(int key, int x, int y)
{
  SoGlutRenderArea *win = SoGlutRenderArea::getComponent(glutGetWindow());
  assert(win != NULL && "glutGetWindow returned invalid window handle");
  win->keyboard_cb(key, false, true, x, y);
}



void SoGlutRenderArea::keyboard_cb(int key, bool down, bool spec, int x, int y)
{
  SoKeyboardEvent::Key tk;

  if (!spec)
    // ascii character
    if (key<128)  tk = keyTranslateTable[key];
    else tk = SoKeyboardEvent::ANY;
  else
    // special key
    if (key >= GLUT_KEY_LEFT)
      // arrows, home, insert, ...
      if (key <= GLUT_KEY_INSERT)  
        tk = specialKeyTranslateTable2[key-GLUT_KEY_LEFT];
      else 
        tk = SoKeyboardEvent::ANY;
    else
      // Fxx keys
      if (key <= GLUT_KEY_F12 && key >= GLUT_KEY_F1)
        tk = specialKeyTranslateTable1[key-GLUT_KEY_F1];
      else 
        tk = SoKeyboardEvent::ANY;

  keyEvent.setKey(tk);
  keyEvent.setState(down ? SoButtonEvent::DOWN : SoButtonEvent::UP);

  makeEvent(&keyEvent, x, y, glutGetModifiers());
}



void glut_mouse_motion_cb(int x, int y)
{
  SoGlutRenderArea *win = SoGlutRenderArea::getComponent(glutGetWindow());
  assert(win != NULL && "glutGetWindow returned invalid window handle");
  win->mouse_motion_cb(x, y);
}



void SoGlutRenderArea::mouse_motion_cb(int x, int y)
{
  // FIXME: 0 is passed to midifiers (instead of glutGetModifiers())
  // because Linux'es glut gives warnings when getting modifiers
  // in mouse motion or passive motion callback. 2002-02-13 PCJohn
  makeEvent(&mouseMotionEvent, x, y, 0);
}



void glut_mouse_button_cb(int button, int state, int x, int y)
{
  SoGlutRenderArea *win = SoGlutRenderArea::getComponent(glutGetWindow());
  assert(win != NULL && "glutGetWindow returned invalid window handle");

  SoMouseButtonEvent::Button b;
  switch (button)
  {
    case GLUT_LEFT_BUTTON: b = SoMouseButtonEvent::BUTTON1; break;
    case GLUT_RIGHT_BUTTON: b = SoMouseButtonEvent::BUTTON2; break;
    case GLUT_MIDDLE_BUTTON: b = SoMouseButtonEvent::BUTTON3; break;
    default: b = SoMouseButtonEvent::ANY;
  }
  win->mouse_button_cb(b, state==GLUT_DOWN, x, y);
}



void SoGlutRenderArea::mouse_button_cb(SoMouseButtonEvent::Button b, bool down, int x, int y)
{
  mouseButtonEvent.setButton(b);
  mouseButtonEvent.setState(down ? SoButtonEvent::DOWN : SoButtonEvent::UP);
  makeEvent(&mouseButtonEvent, x, y, glutGetModifiers());
}



void SoGlutRenderArea::setFullScreen(bool fs)
{
#if SO_GLUT_USE_GAME_MODE

  if (fs) {
    // entering or updating full screen mode

    if (!fullScreen && fullScreenActive) { // can not have two full screen windows
      SoDebugError::post("SoGlutRenderArea::setFullScreen",
                         "Can not create fullscreen window if there is another visible fullscreen window.");
      assert(false && "Can not create fullscreen window if there is another visible fullscreen window.");
      return;
    }
    
    if (this->isVisible()) {
#if 0 // We need not to destroy fullscreen window before resolution change
      // (documentation says), but it make problems with maximized windows
      // of other applications on Windows platform. PCJohn 2002-04-26
      if (!fullScreen && handle > 0)
#else
      if (handle > 0)
#endif
        destroyWindow();
      fullScreen = fs;
      recreateWindow();
    } else
      fullScreen = fs;
  } else

    // leaving full screen mode
    if (fullScreen) {

      // leave fullscreen
      if (handle > 0)
        doLeaveFullScreen();
      
      // update all structures
      fullScreen = false;
      handle = 0;
      
      // if visible -> create window
      if (this->isVisible())
        createWindow();
    }

#else

  if (fullScreen != fs) {
    fullScreen = fs;
    if (handle < 1) {
      recreateWindow();
    } else
      if (fs) {
        position = SbVec2s(glutGet(GLUT_WINDOW_X), glutGet(GLUT_WINDOW_Y));
        glutFullScreen();
      } else {
        glutReshapeWindow(winSize[0], winSize[1]);
        glutPositionWindow(position[0], position[1]);
      }
  }

#endif
}



bool SoGlutRenderArea::doEnterFullScreen()
{
  // generate glut fullscreen string
  char buf[32];

#if 1 // to avoid use of snprintf
  if (fsWidth >= 10240 || fsWidth < 0 || fsHeight >= 7680 || fsHeight < 0 ||
      fsBPP > 64 || fsBPP < 0) {
    SoDebugError::post("SoGlutRenderArea::doEnterFullScreen",
                       "Parameters width: %d, height: %d and bpp: %d are invalid.",
                       fsWidth, fsHeight, fsBPP);
    return false;
  }

  if (fsBPP == 0) sprintf(buf, "%dx%d", fsWidth, fsHeight);
  else sprintf(buf, "%dx%d:%d", fsWidth, fsHeight, fsBPP);

#else
  if (snprintf(buf, 32, "%dx%d:%d", fsWidth, fsHeight, fsBPP) == 32) {
    SoDebugError::post("SoGlutRenderArea::doEnterFullScreen",
                       "setFullScreen function buffer overrun. Check resolution parameters.");
    return false;
  }
#endif

  // enter game mode
  glutGameModeString(buf);

#if 0
  // FIXME: We do not test for GLUT_GAME_MODE_POSSIBLE, becase it looks
  // like that sometimes it is possible to enter game mode even if
  // GLUT_GAME_MODE_POSSIBLE returns false. This is happens on MacOS and
  // I am not sure about Linux. User can manualy perform this action by 
  // isResolutionSupported function.  2002-02-28 PCJohn
  if (!glutGameModeGet(GLUT_GAME_MODE_POSSIBLE)) return false; 
#endif

  if (!glutEnterGameMode()) return false;
  if (!glutGameModeGet(GLUT_GAME_MODE_ACTIVE)) return false;
#if 0 // this line work well on Win32, but it looks unimplemented on Linux
  if (glutGameModeGet(GLUT_GAME_MODE_WIDTH) <= 0) return false;
#endif
  
  fullScreenActive = true;
  handle = glutGetWindow();

#ifdef _WIN32 // works only on WIN32
  int newWidth = glutGameModeGet(GLUT_GAME_MODE_WIDTH);
  int newHeight = glutGameModeGet(GLUT_GAME_MODE_HEIGHT);
  if (fsBPP != 0)
    fsBPP = glutGameModeGet(GLUT_GAME_MODE_PIXEL_DEPTH);

  if (newWidth != fsWidth || newHeight != fsHeight)
    sizeChanged(newWidth, newHeight);
#endif

  // send resize event (in game mode glut do not call resize callback)
  reshape_cb(fsWidth, fsHeight);

  return true;
}



void SoGlutRenderArea::doLeaveFullScreen()
{
  glutLeaveGameMode();
  fullScreenActive = false;
  handle = 0;
}



#define DEPTH_NUM 3
static struct {
  int depth;
  SoGlutRenderArea::Resolution *res;
  int num;
} depthList[DEPTH_NUM] = 
{
  { 16, NULL, 0 },
  { 24, NULL, 0 },
  { 32, NULL, 0 },
};
static int resolutionNum;


#define RESOLUTION_NUM 6
#define DEPTH_NUM 3
#define RESOLUTIONS_DEFAULTS(DEPTH) { \
    { 2048, 1536, DEPTH, false }, { 1600, 1200, DEPTH, false }, { 1280, 1024, DEPTH, false }, \
    { 1024,  768, DEPTH, false }, {  800,  600, DEPTH, false }, {  640,  480, DEPTH, false } }

static struct {
  struct {
    int width;
    int height;
    int bpp;
    bool available;
  } Res[RESOLUTION_NUM];
} ResList[DEPTH_NUM] = {
  { RESOLUTIONS_DEFAULTS(16) },
  { RESOLUTIONS_DEFAULTS(24) },
  { RESOLUTIONS_DEFAULTS(32) },
};



static void updateResolutionList()
{
  if (!availableRes) {
    resolutionNum = 0;
    int i,j;
    for (j=0; j<DEPTH_NUM; j++) {
      for (i=0; i<RESOLUTION_NUM; i++) {
        bool a = SoGlutRenderArea::isResolutionSupported(ResList[j].Res[i].width, 
            ResList[j].Res[i].height, ResList[j].Res[i].bpp);
        ResList[j].Res[i].available = a;
        if (a)  resolutionNum++;
      }
    }

    if (resolutionNum == 0) {
      availableRes = NULL;
      for (int i=0; i<DEPTH_NUM; i++) {
        depthList[i].res = NULL;
        depthList[i].num = 0;
      }
      return;
    }

    availableRes = (SoGlutRenderArea::Resolution*)malloc(resolutionNum*sizeof(SoGlutRenderArea::Resolution));

    int n = 0;
    for (j=0; j<DEPTH_NUM; j++) {
      for (i=0; i<RESOLUTION_NUM; i++) {
        if (ResList[j].Res[i].available) {
          availableRes[n].width  = ResList[j].Res[i].width; 
          availableRes[n].height = ResList[j].Res[i].height; 
          availableRes[n].bpp    = ResList[j].Res[i].bpp;
          n++;
        }
      }
    }
  }
}



int SoGlutRenderArea::getResolutionList(const Resolution *&resList, int depthBits)
{
  updateResolutionList();

  if (depthBits == 0) {
    resList = availableRes;
    return resolutionNum;
  } else {
    int i;
    for (i=0; i<DEPTH_NUM; i++) {
      if (depthList[i].depth == depthBits) {
        resList = depthList[i].res;
        return depthList[i].num;
      }
    }
    resList = NULL;
    return 0;
  }
}



bool SoGlutRenderArea::isResolutionSupported(int width, int height, int bpp)
{
  char buf[32];
#if 1 // to avoid use of snprintf
  if (width >= 10240 || width < 0 || height >= 7680 || height < 0 ||
      bpp > 64 || bpp < 0)  return false;
  sprintf(buf, "%dx%d:%d", width, height, bpp);
#else
  if (snprintf(buf, 32, "%dx%d:%d", width, height, bpp) == 32) {
    SoGlut::warning("isResolutionSupported function buffer overrun. Check passing values.", NULL);
    return false;
  }
#endif
  glutGameModeString(buf);
  return (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE) != 0);
}



void SoGlutRenderArea::centerWindow()
{
  int screenWidth = glutGet(GLUT_SCREEN_WIDTH);
  int screenHeight = glutGet(GLUT_SCREEN_HEIGHT);
  SbVec2s dim = getSize();
  int left = (screenWidth - dim[0]) >> 1;
  int top = (screenHeight - dim[1]) >> 1;
  setPosition(SbVec2s(left, top));
}



